package MSN;

# A module to allow msn stuff to happen in perl :-)
# (C)opyright Mark Zealey <mark@zealos.org> 2002
use strict;
use Digest::MD5 qw/md5_hex/;
use IO::Select;
use IO::Socket::INET;

# Use high-res timing if avaliable
BEGIN { eval 'use Time::HiRes qw/time/' }

# The usual msn ports
use constant DEFAULT_MSN_PORT => 1863;
use constant DEFAULT_FTP_PORT => 6891;

# The standard header we use to send a message
our $MSN_STD_HDR = <<EOF;
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
X-MMS-IM-Format: FN=MS%20Sans%20Serif; EF=; CO=0; CS=0; PF=0

EOF

# The standard header for file transfer communication
our $MSN_FILE_HDR = <<EOF;
MIME-Version: 1.0
Content-Type: text/x-msmsgsinvite; charset=UTF-8

EOF

# Version string we send to server on connection
our $MSN_VERS = 'MSNP7 MSNP6 MSNP5 MSNP4 CVR0';

# The magic we add to a CHL response. Looks like M$ tried to kill us *sigh*
our $MSN_QRY_MAGIC = 'Q1P7W2E4J9R8U3S5';

# Hash of all the different commands we can send to it.
our %MSNCMD = (
		version => 'VER',
		ack => 'ACK',
		info => 'INF',
		change_servers => 'XFR',
		user => 'USR',
		sync => 'SYN',
		change_fuser => 'REA',
		change_my_state => 'CHG',
		quit => 'BYE',
		join => 'JOI',
		message => 'MSG',
		list => 'LST',
		swb_list => 'IRO',
		add => 'ADD',
		del => 'REM',
		ring => 'RNG',
		answer => 'ANS',
		initial_state => 'ILN',
		offline => 'FLN',
		change_state => 'NLN',
		invite => 'CAL',
		out => 'OUT',
		remove => 'REM',
		size => 'FIL',
		transfer => 'TFR',
		challange => 'CHL',
		query => 'QRY',
		ping => 'PNG',
		pong => 'QNG',		# Neat ;)

		# Errors (as numbers)
		error_list_clash => '216',
		error_auth_failed => '911',
	);

# Hash to convert different states to messanger format
our %STATES = (
		online => 'NLN',
		offline => 'FLN',
		busy => 'BSY',
		idle => 'IDL',
		hidden => 'HDN',
		brb => 'BRB',
		phone => 'PHN',
		lunch => 'LUN',
		away => 'AWY',
	);

our %OUTS = (
		OTH => 'You have logged in from another computer',
		SSD => 'Server going down for an upgrade',
	);

our %REV_STATES = reverse %STATES;

sub new {
	my $class = shift;
	$class = ref $class || $class;

	my $self = {
			@_,
			session => 0,	# Main connection ID
			rings => [],
			events => {},
			files => [],
			state => 'Not connected',
			buffer => '',
		};

	$self->{Debug} ||= undef;
	defined $self->{Debug} and $self->{Debug} == 1 and $self->{Debug} =
								\*STDERR;
	
	bless $self, $class;
	$self->{MSN} = $self;
}

sub add {
	my $self = shift;
	%$self = (%$self, @_)
}

sub connected {
	my $self = shift;
	$self and $self->{sock} and $self->{sock}->connected
}

sub con {
	my $self = shift;

	my ($host, $port) = $self->{server} =~ /^([^:]+)(?::(\d+))?$/
					or return $self->err('Invalid server');

	$self->{sock} = IO::Socket::INET->new(PeerAddr => "$host:$port") or
						return $self->err("Error: $!");

	# Attempt to set socket non-blocking
	if(!defined $self->{sock}->blocking(0)) {
		$self->err("Error with setting blocking: $!");
		undef $self->{sock};
		return undef
	}

	1
}

sub connect {
	my $self = shift;

	%$self = (%$self, @_);

	$self->{state} = 'connecting';

	$self->{Host} ||= 'messenger.hotmail.com';
	$self->{Port} ||= DEFAULT_MSN_PORT;

	return $self->err("No username defined") unless $self->{User};
	return $self->err("No password defined") unless $self->{Pass};

	# Fixup the username for a lazy luser
	$self->{User} = $self->fix_user($self->{User});
	$self->{FUser} ||= "I'm too lame to change my fusername";

	$self->{server} = $self->{Server} || "$self->{Host}:$self->{Port}";
	$self->{users} = {};

	$self->debug("Connecting to $self->{server}");

	return $self->err unless $self->con;

	$self->add_fh($self->{sock}, $self);

	# Send version
	$self->send_line('version', $MSN_VERS) or return $self->err;

	1
}

# Change your fuser-name aka 'screen name'
sub change_fuser {
	my ($self, $name) = @_;

	$self = $self->{MSN};
	$name ||= $self->{FUser};
	$name =~ s/([^\w\d])/sprintf '%%%2x', ord $1/eg;

	$self->send_line('change_fuser', $self->{User}, $name)
}

sub change_my_state {
	my ($self, $state) = @_;

	exists $STATES{lc $state} or return undef;

	# Already in this state? Don't waste bw :-)
	exists $self->{state} and lc $self->{state} eq lc $state and return 1;

	$self->send_line('change_my_state', $STATES{lc $state});
}

sub get_state { shift->{MSN}{state} }

sub disconnect {
	my $self = shift;

	$self->{state} = 'Not connected';

	# Kill all our ring connections
	$_->disconnect for grep {defined} @{$self->{rings}};

	if($self->connected) {
		$self->send_line('out', @_);
		close $self->{sock}
	}

	delete $self->{sock};
	delete $self->{select};
	1
}

sub get_fhs {
	my $self = shift;

	for(@{$self->{rings}}, @{$self->{FTPs}}) {
		if(defined and !defined $_->{sock} and defined $_->{sock_fn}) {
			if(exists $_->{session}) {		# Switchboard
				$self->debug('Closing switchboard ' .
								$_->{session})
			} else {				# Transfer
				$self->debug('Closing FTP transfer ' .
								$_->{Num});
				delete $self->{_FTPCookie}{$_->{Cookie}}
			}
			$self->del_fh($_->{sock_fn});
			undef $_
		}
	}

	keys %{$self->{fh_hash}}
}

sub from_fileno {
	my ($self, $fileno) = @_;
	$self->{fh_hash}{$fileno}
}

sub add_fh {
	my ($self, $fh, $obj) = @_;

	my $fileno = $fh;
	# If we were passed a fileno, don't fileno it :)
	$fileno = fileno $fh if ref $fh;

	$self->{fh_hash}{$fileno} = $obj
}

sub del_fh {
	my ($self, $fh) = @_;

	my $fileno = $fh;
	# If we were passed a fileno, don't fileno it :)
	$fileno = fileno $fh if ref $fh;

	delete $self->{fh_hash}{$fileno}
}

# See if there's any stuff we need to handle and handle it.
#
# Optionally pass a variable saying if you want timeout to it and then a load of
# IO::Select objects and it will return those if any of them have been selected.
sub do_stuff {
	my ($self, $loop, $timeout) = (shift, shift, shift);

	# Build a list of our user-specified fh's
	my %user_handles = map {fileno($_) => $_} @_;
	# Anything to do?
	do {{
		my @ret;

		# Not connected, nothing to do..
		return $self->call_callback('disconnected') unless $self->connected;

		# Regenerate the list of handles each time as they may change
		# between loops.
		my $handles = IO::Select->new(keys %user_handles,
							$self->get_fhs);

		# List of our handles is now in the IO::Select object, select it
		my @a = IO::Select->select($handles, (undef) x 2, $timeout);

		# If there was an error with that, it could have been because of
		# an interrupt or something, quietly exit from this.
		last unless @a;

		for(@{$a[0]}) {
			if(exists $user_handles{$_}) {		# Is a luser-fh?
				push @ret, $user_handles{$_};
				next
			}

			my $t = $self->from_fileno($_);

			$t->get_block and $t->internal_do_stuff or
								$t->disconnect
		}

		return [@ret] if @ret
	}} while($loop);

	[]
}

sub internal_do_stuff {{
	my $self = shift;

	if($self->{msg_wait}) {			# Waiting for a message
		return 1 if length $self->{buffer} < $self->{msg_wait}[0];

		# Got a full message, handle it in handle_msg()
		my $msg = substr $self->{buffer}, 0, $self->{msg_wait}[0], '';
		$self->handle_msg(@{$self->{msg_wait}}[1,2], $msg);

		undef $self->{msg_wait}
	}

	# o_O we have something to do!
	my $line = $self->get_line or return 1;

	my @bit = split ' ', $line;

	# First check if this event is one someone else wants to catch
	if($bit[1] and $bit[1] =~ /^\d+$/ and exists $self->{events}{$bit[1]}) {
		return $self->call_del_event($bit[1], $self, $line)
	}

	# Else, look through our handlers
	if($bit[0] eq $MSNCMD{change_servers}) {
		$self->change_server(Server => $bit[3])
	} elsif($bit[0] eq $MSNCMD{message}) {	# We got a message
		# Message format, USER LENGTH where LENGTH is the length
		# to read in bytes, neat
		$self->{msg_wait} = [@bit[3,1,2]]
	} elsif($bit[0] eq $MSNCMD{list}) {	# List maintainance
		$self->add_user(@bit[2,6,7]) if defined $bit[7]
	} elsif($bit[0] eq $MSNCMD{add}) {
		# Index here == 0...
		my $u = $self->add_user(@bit[2,4,5]);

		if($bit[2] eq 'AL') {
			$self->call_callback('auth_success', $u)
		} elsif($bit[2] eq 'BL') {
			$self->call_callback('block_success', $u)
		} elsif($bit[2] eq 'FL') {
			$self->call_callback('add_success', $u)
		}
	} elsif($bit[0] eq $MSNCMD{del}) {
		my $u = $self->del_user(@bit[4,2]);

		if($bit[2] eq 'AL') {
			$self->call_callback('deauth_success', $u);
		} elsif($bit[2] eq 'BL') {
			$self->call_callback('unblock_success', $u);
		} elsif($bit[2] eq 'FL') {
			$self->call_callback('del_success', $u);
		} elsif($bit[2] eq 'RL') {
			$self->call_callback('del', $u);
		}
	} elsif($bit[0] eq $MSNCMD{ring}) {	# Someone is ringing us
		my $swb = MSN::Switchboard->new(
						ID => $bit[1],
						Server => $bit[2],
						Auth => $bit[4],
						MSN => $self,
					) or return;

		$self->{rings}[$swb->{session}] = $swb
	} elsif($bit[0] eq $MSNCMD{join}) {
		# Add the user
		$self->_join_user(@bit[1,2]);
	} elsif($bit[0] eq $MSNCMD{quit}) {
		$self->leave($bit[1], $self->get_sess_id);
	} elsif($bit[0] eq $MSNCMD{out}) {
		$self->call_callback('out', $OUTS{$bit[1]});
		$self->disconnect;
		return $self->err("OUT recieved: $bit[1]")
	} elsif($bit[0] eq $MSNCMD{change_my_state}) {
		$self->{state} = $REV_STATES{$bit[2]};
		$self->call_callback('state_success', $bit[2])
	} elsif($bit[0] eq $MSNCMD{change_fuser}) {
		return $self->err unless $bit[3] eq $self->{User};

		$self->{FUser} = $bit[4];

		$self->call_callback('fuser_success', $self->{FUser})
	} elsif($bit[0] eq $MSNCMD{challange}) {
		$self->send_line('query', 'msmsgs@msnmsgr.com 32') or
							return $self->err;
		print {$self->{sock}} md5_hex($bit[2] . $MSN_QRY_MAGIC);
	} elsif($bit[0] eq $MSNCMD{pong}) {
		my $sent = shift @{$self->{pings}};
		$sent or return $self->err;

		$self->call_callback('ping', time - $sent)

	# With these three, we just change the user's state, dunno why the
	# protocol didn't originally do this...
	} elsif($bit[0] eq $MSNCMD{initial_state}) {
		$self->change_user_state(@bit[2..4])
	} elsif($bit[0] eq $MSNCMD{change_state}) {
		$self->change_user_state(@bit[1..3])
	} elsif($bit[0] eq $MSNCMD{offline}) {
		# First, disconnect the user from the swb sessions. (Some buggy
		# clients don't BYE from the swb sessions when they quit...)
		my $u = $self->get_user($bit[1]);

		$self->leave($bit[1], $_) for grep /^\d+$/, keys %{$u->{Lists}};

		# And change the users state to offline
		$self->change_user_state($STATES{offline}, $bit[1])

	# Error handling
	} elsif($bit[0] eq $MSNCMD{error_auth_failed}) {
		$self->err('Bad Password');
		$self->call_callback('auth_failed');
		return $self->disconnect
	} elsif($bit[0] =~ /^\d{3}$/) {		# Usially means an error
		$self->err("Unknown error number: $bit[0]")

	# When joining a switchboard:
	} elsif($bit[0] eq $MSNCMD{swb_list}) {
		$self->_join_user(@bit[4,5]);
	} elsif($bit[0] eq $MSNCMD{answer}) {
		# Ignore it.

	# These next ones are only used in the initalization process:
	} elsif($bit[0] eq $MSNCMD{version}) {
		if($bit[2] eq '0') {
			$self->disconnect;

			return $self->err('Protocol not supported')
		}

		$self->send_line('info', '');

	# Server responds with the auth protocol we use
	} elsif($bit[0] eq $MSNCMD{info}) {
		return $self->err("Auth protocol $bit[2] not supported")
							if $bit[2] ne 'MD5';

		# ACK it and send username
		$self->send_line('user', "MD5 I $self->{User}")
							or return $self->err;
	} elsif($bit[0] eq $MSNCMD{user}) {
		if($bit[2] eq 'MD5') {		# Send pword
			$self->send_line('user', 'MD5 S',
					md5_hex("$bit[4]$self->{Pass}"))
							or return $self->err
		} elsif($bit[2] eq 'OK') {	# Whee, auth'd properly
			return 1 if $self->get_sess_id;		# Don't on swb's
			$self->change_fuser or return $self->err;
			$self->sync or return $self->err;

			# Announce to the world that we're around!
			$self->change_my_state('online') or return $self->err;

			$self->call_callback('connect');
		} else {
			$self->debug("Not sure about '$line'");
		}
	} else {
		$self->debug("Not sure about '$line'");
	}

	push @_, $self;				# XXX: KLUGE!! :(
	redo
}}

# Sync says what the status of the local lists are.. as we don't have local
# lists, we just send 1 so it gives us everything =)
sub sync { shift->send_line('sync', 1) }

sub ping {
	my $self = shift;
	$self->send_line('ping');
	push @{$self->{pings}}, time;
}

# A user has left a switchboard session
sub leave {
	my ($self, $user, $sess) = @_;

	my $u = $self->del_user($user, $sess);

	my $swb = $self->{rings}[$sess];

	return unless defined $swb;

	delete $swb->{User}{$user};

	$swb->call_callback('part', $u);

	# Disconnect unless we have more people here.
	$swb->disconnect unless keys %{$swb->{User}}
}

# Change the state of a user.
sub change_user_state {
	my ($self, $state, $user, $fuser) = @_;

	my $u = $self->get_user($user, $fuser) or return
						$self->unknown_user($user);

	# Return if the states are equal
	return 1 if $u->{State} eq $state;

	$u->{State} = $state;

	$self->call_callback('change_state', $u)
}

sub start_ring { shift->_start_ring(undef, @_) }

sub _start_ring {
	my $self = shift;

	# Get info about a new switchboard to connect to
	my $id = $self->send_line('change_servers', 'SB') or return $self->err;

	$self->add_event($id, \&_ring_response, @_)
}

sub _ring_response {
	my ($self, $line, $ftp, @users) = @_;

	my @bit = split ' ', $line;

	my $swb = MSN::Switchboard->new(
					Server => $bit[3],
					Auth => $bit[5],
					MSN => $self,
				) or return $self->err;

	$self->{rings}[$swb->get_sess_id] = $swb;

	$swb->invite_users(undef, @users);		# Invite the users
}

sub invite_users {
	my ($self, $swb_num) = (shift, shift);

	my $swb = $self;
	$swb = $self->{rings}[$swb_num] if $swb_num;

	return $self->err unless defined $swb;

	# Invite the users to the switchboard now
	$swb->_invite_user($_) for @_;
}

# Return the state in a human-readable format
sub get_human_state {
	my ($self, $user) = @_;
	my $u = $self->get_user($user) or return $self->unknown_user($user);
	my $state = $u->{State};
	return $REV_STATES{$state} if exists $REV_STATES{$state};
	'Unknown'
}

sub handle_msg {
	my ($self, $user, $fuser, $msg) = @_;
	my %message;

	# No fuser needed, msn keeps the same fuser when talking to you even if
	# the luser has changed it. Silly Billy...
	$self->get_user($user);

	$message{User} = $user;

	($message{Raw} = $msg) =~ s/\r//g;

	# With any luck, we got here with no errors.
	($message{MIMEVer}) = $message{Raw} =~ /MIME-Version: (.*)\n/i;

	$self->debug("Mime version $message{MIMEVer} not supported fully")
				unless $message{MIMEVer} == 1;

	($message{Type}) = $message{Raw} =~ /Content-Type: (.*?)[;\n]/i;

	if($message{Type} eq 'text/plain') {
		# A normal message
		($message{Body}) = $message{Raw} =~ /\n\n(.*?)$/is;
		$message{Body} ||= '';
#		$self->debug(map { ord } split //, $message{Body});
		return $self->call_callback('message', \%message);
	} elsif($message{Type} eq 'application/x-msmsgsemailnotification') {
		# An email message - Added by reggy
		my $t = 0;
		($message{From}) = $message{Raw} =~ /From: (.*)/i or $t = 1;
		($message{Subject}) = $message{Raw} =~ /Subject: (.*)/i
								or $t = 1;
		($message{FromAddr}) = $message{Raw} =~ /From-Addr: (.*)/i
								or $t = 1;

		return $self->call_callback('new_mail', \%message) if $t == 0;
	} elsif($message{Type} eq 'text/x-msmsgscontrol') {
		return $self->call_callback('user_types', \%message)
	} elsif($message{Type} eq 'text/x-msmsgsinvite') {	# File transfer
		for(split /\n/, $message{Raw}) {
			/^([^:]+):\s*(.*)$/ and $message{$1} = $2
		}

		if(exists $message{'Application-Name'} and
			$message{'Application-Name'} eq 'voice conversation') {
			$self->debug('Voice not supported. (yet :)');
			$self->debug($message{Raw});
			return
		}

		# Abreviate these..
		my %t = (	'Application-FileSize' => 'Size',
				'Application-File' => 'File',
				'IP-Address' => 'Host',
				'Invitation-Cookie' => 'Cookie',
				'AuthCookie' => 'Auth',
			);

		for(keys %t) {
			$message{$t{$_}} = $message{$_} if exists $message{$_}
		}
		
		$message{Sess} = $self->get_sess_id;

		if($message{'Invitation-Command'} eq 'INVITE') {
			my $ftp = MSN::FTP::Get->new (
							MSN => $self->{MSN},
							SWB => $self,
							%message
						) or return $self->err;

			return $self->{MSN}->call_callback('file_invite', $ftp);
		} elsif($message{'Invitation-Command'} eq 'ACCEPT') {
			my $ftp = $self->{MSN}->file_req_find($message{Cookie});
			$ftp->add(%message);

			$self->{MSN}->call_callback('file_start', $ftp);

			# We can start downloading now!
			return $ftp->connect;
		}
	}

	$self->debug("Message type $message{Type} not supported");
	$self->debug($message{Raw});
}

sub file_get {
	my ($self, $num, $get) = @_;

	my $message = $self->{MSN}->file_req($num) or return $self->err;
	

	my $s = 'ACCEPT';
	$s = 'REJECT' unless $get;

	$message->{SWB}->send_std_msg($MSN_FILE_HDR, <<EOF);
Invitation-Command: $s
Invitation-Cookie: $message->{Cookie}
EOF
}

sub file_send {
	my ($self, $user, $file, $size) = @_;

	my $fh;

	if(!ref $file) {
		open $fh, $file or return $self->err("Can't open $file, $!");
		$size ||= -s $file
	}
	$fh = $file if ref $file eq 'GLOB';

	return $self->err("Not being passed a valid file ref: $file") if !$fh;
	return $self->err("Size $size bytes is invalid") if !$size or $size < 1;
	
	# Always create a new swb first..
	$self->_start_ring(	{
					FH => $fh,
					Size => $size,
				}, $user
			);
}

# Adds the specified message's info to the file download list
sub file_req_add {
	my ($self, $ftp) = @_;

	my $i = 0;

	$i++ while $self->{FTPs}[$_];

	$self->{FTPs}[$i] = $ftp;
	$self->{_FTPCookie}{$ftp->{Cookie}} = $ftp;

	$i
}

sub file_req_find { shift->{_FTPCookie}{+shift} }

sub file_req { shift->{FTPs}[shift] }

sub send_msg_swb {
	my ($self, $swbnum) = (shift, shift);
	return undef unless $self->{rings}[$swbnum];

	$self->{rings}[$swbnum]->send_msg(@_)
}

sub send_msg { shift->send_std_msg($MSN_STD_HDR, @_) }

sub send_std_msg {
	my ($self, $msg) = (shift, shift);

	$msg .= join ' ', @_;

	$msg =~ s/\r?\n/\r\n/g;			# Update line-endings

	$self->send_line('message', 'N', length $msg) or return $self->err;

	print {$self->{sock}} $msg or return $self->err;
}

# Add an event that will callback the specified function
sub add_event {
	my ($self, $id, $callback, @args) = @_;

	$self->{events}{$id} = [$callback, [@args]];
}

# Remove event, call and return
sub call_del_event {
	my ($self, $id, @args) = @_;

	my $t = $self->{events}{$id};

	&{$t->[0]}(@args, @{$t->[1]});
	delete $self->{events}{$id};
}

sub sess_exists {
	my ($self, $num) = @_;
	exists $self->{rings}[$num] and defined $self->{rings}[$num]
}

sub part_sess {
	my ($self, $num) = (shift, shift);
	return $self->err("Session $num non-existant")
						unless $self->sess_exists($num);

	$self->{rings}[$num]->disconnect(@_)
}

# Get a unique session ID
sub get_usid {
	my $self = shift;

	my $i = 1;

	$i++ while $self->{rings}[$i];

	$i
}

# If the switchboard session has ended, this is called.
sub remove_swb_sess { undef shift->{rings}[shift()] }

sub get_sess_id { shift->{session} }

sub add_callback {
	my ($self, $name, $func) = @_;
	push @{$self->{callbacks}{$name}}, $func;
}

sub del_callback {
	my ($self, $name, $func) = @_;

	if(defined $func) {
		@{$self->{callbacks}{$name}} =
				grep $_ ne $func, @{$self->{callbacks}{$name}}
	} else {
		delete $self->{callbacks}{$name}
	}
}

sub call_callback {
	my ($self, $name) = (shift, shift);

	for(@{$self->{callbacks}{$name}}) {
		eval { &{$_}($self, @_) };
		$self->debug("Hmm, function didn't work, $@") if $@
	}
}

# AL -> Allow List (Ones who are friends with me)
# BL -> Ban List (My enemys)
# FL -> Friends List (Ones I'm friends with)
# RL -> Request List (Ones that want to be my friend)

sub _user_is_in_list {
	my ($self, $list, $user, $fuser) = @_;
	my $u = $self->get_user($user, $fuser) or return
						$self->unknown_user($user);

	exists $u->{Lists}{$list} 
}

sub user_is_allowed { shift->_user_is_in_list('AL', @_) }
sub user_is_banned { shift->_user_is_in_list('BL', @_) }
sub user_is_friend { shift->_user_is_in_list('FL', @_) }
sub user_is_request { shift->_user_is_in_list('RL', @_) }

# Fix a username
sub fix_user {
	my ($self, $user) = @_;

	$user =~ /^([^@]+)(.*)/;

	lc $1 . ($2 || '@hotmail.com')
}

# Return a plain-text decoded fuser
sub plain_fuser {
	my ($self, $fuser) = @_;
	$fuser =~ s/\%([0-9A-Fa-f]{2})/chr hex $1/eg;
	$fuser
}

# list name, name, friendly name, (optional) initial state
sub add_user {
	my ($self, $list, $user, $fuser, $state) = @_;

	$user = $self->fix_user($user);

	if(exists $self->{users}{$user}) {		# Add list in
		$self->{users}{$user}{Lists}{$list} = 1
	} else {					# Need to create new one
		($fuser) = $user =~ /^([^@]+)/ if !defined $fuser;

		$self->{users}{$user} = {
					User => $user,
					FUser => $fuser,
					State => $state || $STATES{offline},
					Lists => {
						$list => 1
					},
				};
	}

	# If someone wants to be added, isn't already there and
	# isn't in our banned list, call the callback
	if($list eq 'RL' and !$self->user_is_allowed($user)
			and !$self->user_is_banned($user)) {
		$self->call_callback('add', $self->get_user($user));
	}

	$self->{users}{$user}
}

sub del_user {
	my ($self, $user, $list) = @_;

	my $u = $self->get_user($user) or return $self->unknown_user($user);

	if(defined $list) {
		$self->debug("Removing list $list from user $user");
		delete $u->{Lists}{$list};
	} else {
		# No list specified
		delete $self->{users}{$u->{User}}
	}

	$u
}

# Same here for fuser as in add_user.
sub get_user {
	my ($self, $user, $fuser) = @_;

	$user = $self->fix_user($user);

	if(defined $fuser and $self->{users}{$user}{FUser} ne $fuser) {
		my $old_fuser = $self->{users}{$user}{FUser};
		$self->{users}{$user}{FUser} = $fuser;
		$self->{MSN}->call_callback('fuser_change',
					$self->get_user($user), $old_fuser)
	}

	$self->{users}{$user}
}

sub unknown_user {
	my ($self, $user) = @_;
	$self->call_callback('unknown_user', $user);
	1
}

sub do_contact {
	my ($self, $list, $user, $fuser) = @_;

	my $u = $self->get_user($user, $fuser) || $self->add_user($list,
								$user, $fuser);

	$self->send_line('add', "$list $u->{User} $u->{FUser}") or
							return $self->err;
	$self->add_user($list, $user);

	$u
}

sub rem_contact {
	my ($self, $list, $user) = @_;

	my $u = $self->get_user($user) or return $self->unknown_user($user);

	$self->send_line('del', $list, $u->{User}) or return $self->err;
	$self->del_user($user, $list);
}

sub authorize_contact	{ shift->do_contact('AL', @_) }
sub deauthorize_contact	{ shift->rem_contact('AL', @_) }
sub add_contact		{ shift->do_contact('FL', @_) }
sub del_contact		{ shift->rem_contact('FL', @_) }

sub block_contact { 
	my $self = shift;
	$self->rem_contact('AL', @_);
	$self->do_contact('BL', @_);
}

sub unblock_contact { 
	my $self = shift;
	$self->rem_contact('BL', @_);
	$self->do_contact('AL', @_);
}

# Disconnects and reconnects to the server in Host and Port.
sub change_server {
	my $self = shift;

	$self->disconnect or return $self->err;

	%$self = (%$self, @_);

	$self->connect
}

# Sends a line to the server, returns undef on failure, the contents of the line
# it sent otherwise.
sub send_line {
	my ($self, $str, @text) = @_;

	return $self->err("$str doesn't exist!") unless $MSNCMD{$str};
	if(!$self->connected) {
		$self->call_callback('disconnected');
		return undef;
	}

	my $send = $MSNCMD{$str};
	
	$send .= ' ' . ++$self->{trans} . " @text" if @text;

	$self->debug("> $send");

	if(!print {$self->{sock}} "$send\r\n") {	# Disconnected
		$self->call_callback('disconnected');
		return undef;
	}

	$self->{trans}
}

# Reads anything that can be read into the buffer. Returns undef on error.
sub get_block {
	my $self = shift;

	return $self->err('Not connected to server') unless $self->connected;

	my $t;
	my $ret = read $self->{sock}, $t, 2**20;	# Some large value..

	if(!defined $ret and $! !~ /temporarily unav/i) {	# Eweee.. error!
		$self->debug('Remote read error on session ' .
						$self->get_sess_id . ": $!");

		$self->disconnect;

		return undef
	}

	# shove the rest into the buffer and see if we can read a line from it
	$self->{buffer} .= $t if $ret;

	1
}

# Returns the string got or undef on error/not a full string
sub get_line {
	my $self = shift;

	$self->{buffer} =~ s/(.*?)\r?\n//;
	$self->debug("< $1") if $1;

	$1
}

sub debug {
	my ($self, $t) = (shift, join(' ', @_));
	print {$self->{Debug}} "$self->{session} $t\n" if $self->{Debug};
}

sub err {
	my ($self, $text) = (shift, join(' ', @_));

	if($text eq '') {
		my ($package, $file, $line) = caller;
		$text = "Unknown error at line $line, file $file"
	}

	$self->{Error} = $text;

	$self->debug("Error: $text");
	undef
}

package MSN::Switchboard;

our @ISA = qw(MSN);

sub new {
	my $class = shift;
	$class = ref $class || $class;

	my $self = {
			@_,
			buffer => '',
		};

	bless $self, $class;

	$self->{session}	= $self->{MSN}->get_usid;
	$self->{Debug}		||= $self->{MSN}{Debug};
	$self->{Me}		||= $self->{MSN}{User};
	$self->{callbacks}	= $self->{MSN}{callbacks};
	$self->{users}		= $self->{MSN}{users};
	$self->{rings}		= $self->{MSN}{rings};

	$self->{server} = $self->{Server} || "$self->{Host}:$self->{Port}";

	$self->debug("Connecting switchboard to $self->{server}");

	return $self->err unless $self->con;
	$self->{sock_fn} = fileno $self->{sock};

	$self->{MSN}->add_fh($self->{sock_fn}, $self);

	# We opened successfully, call the open_chat callback
	$self->call_callback('open_chat', $self->get_sess_id);

	if(exists $self->{ID}) {		# We've been invited here
		# We're now connected, auth ourselves
		$self->send_line('answer', @{%$self}{qw/Me Auth ID/})
							or return $self->err;
	} else {				# We set this chat up
		# Auth ourselves
		$self->send_line('user', @{%$self}{qw/Me Auth/})
							or return $self->err;
	}

	$self
}

sub disconnect {
	my $self = shift;

	# Remove users that might have been involved in the chat
	$self->del_user($_, $self->get_sess_id) for keys %{$self->{User}};
	$self->call_callback('close_chat', $self->get_sess_id);

	if($self->connected) {
		$self->send_line('out');
		close $self->{sock}
	}

	undef $self->{sock};
	1
}

sub _invite_user {
	my ($self, $user) = @_;

	my $u = $self->get_user($user) or return $self->unknown_user($user);

	$self->send_line('invite', $u->{User}) or return $self->err;
}

sub _join_user {
	my ($self, $user, $fuser) = @_;

	# Update fname etc
	my $u = $self->add_user($self->get_sess_id, $user, $fuser);
	$self->{User}{$u->{User}} = 1;

	$self->call_callback('join', $u)
}


package MSN::FTP;

# We inherit from MSN
our @ISA = qw(MSN);

our $MSN_FTP_VER = 'MSNFTP';

sub disconnect {
	my $self = shift;

	close $self->{sock} if defined $self->{sock};
	undef $self->{sock};

	1
}

sub debug {
	my $self = shift;
	print {$self->{Debug}} "FTP$self->{Num} @_\n" if $self->{Debug};
}

sub send_line {
	my ($self, $str, @text) = @_;

	return $self->err('Not connected to FTP') unless $self->connected;
	return $self->err("$str doesn't exist!") unless $MSNCMD{$str};
	if(!$self->connected) {
		$self->call_callback('disconnected');
		return undef;
	}

	my $send = $MSNCMD{$str};
	
	$send .= " @text" if @text;

	$self->debug("> $send");

	if(!print {$self->{sock}} "$send\r\n") {	# Disconnected
		$self->call_callback('disconnected');
		return undef;
	}
}


package MSN::FTP::Get;

our @ISA = qw(MSN::FTP);

sub new {
	my $class = shift;
	$class = ref $class || $class;

	my $self = {
			State => 'Waiting',
			@_,
			buffer => '',
		};

	bless $self, $class;

	return $self->err unless exists $self->{MSN};

	$self->{Num} = $self->{MSN}->file_req_add($self);
	$self->{Debug} ||= $self->{MSN}{Debug};

	$self
}

sub connect {
	my $self = shift;

	%$self = (%$self, @_);

	$self->{State} = 'Connecting';

	$self->{downloading} = 0;		# Not downloading yet

	open $self->{FH}, '>', $self->{File} or return
			$self->err("Can't open $self->{File} for writing, $!");

	$self->{server} = $self->{Server} || "$self->{Host}:$self->{Port}";

	$self->con or return $self->err("Couldn't connect");
	$self->{sock_fn} = fileno $self->{sock};

	$self->{MSN}->add_fh($self->{sock_fn}, $self);

	$self->send_line('version', $MSN_FTP_VER) or return $self->err;
}

sub internal_do_stuff {
	my $self = shift;

	if($self->{downloading}) {{	# Get a block, write to file
		# 3 bytes is header:
		# Byte number | Meaning
		# 0           | unknown
		# 1           | Number of bytes in this chunk (lower byte)
		# 2           | Number of bytes in this chunk (higher byte)

		if(!$self->{next_chunk}) {
			return 1 unless length $self->{buffer} >= 3;
			my $t = substr $self->{buffer}, 0, 3, '';
			my @t = map {ord} split //, $t;
			$self->{next_chunk} = $t[2] * 256 + $t[1];
			$self->debug(@t, $self->{next_chunk});
		}

		return 1 unless length $self->{buffer} >= $self->{next_chunk};

		my $in = substr $self->{buffer}, 0, $self->{next_chunk}, '';

		print {$self->{FH}} $in;

		$self->{to_read} -= length $in;
		$self->debug("To read: $self->{to_read}");

		if(!$self->{to_read}) {			# Nothing left to read
			close $self->{FH};

			# Huh? I wonder what the point of the magic number is..
			# can't figure it out, it appears in all the transfer
			# dumps I've seen.. anyone got any ideas?
			$self->send_line('quit', 16777989);
			return $self->disconnect
		}

		print {$self->{sock}} chr(0) x 6 if $self->connected;	# ACK
		undef $self->{next_chunk};

		redo
	}}

	while(defined(my $line = $self->get_line)) {
		my @bit = split ' ', $line;

		if($bit[0] eq $MSNCMD{version}) {
			if($line !~ /$MSN_FTP_VER/) {
				return $self->err('Incompatable protocols:',
									$line)
			}

			$self->send_line('user', $self->{MSN}{User},
					$self->{Auth}) or return $self->err;
		} elsif($bit[0] eq $MSNCMD{size}) {
			$self->{Size} = $bit[1];
			$self->{to_read} = $self->{Size};
			$self->{downloading} = 1;
			$self->send_line('transfer')	# Let it rip boyz
		}
	}

	1
}

1
